import logging
from pathlib import Path

from storage import UploadedEntity

from portal.database import Project, Submission, SubmissionState
from portal.service import general

log = logging.getLogger(__name__)


class SubmissionProcessor:
    def __init__(self, submission: Submission, params: dict = None):
        self._submission = submission
        self._params = params

    @property
    def submission(self) -> Submission:
        return self._submission

    @property
    def params(self) -> dict:
        return self._params

    @property
    def project(self) -> Project:
        return self.submission.project

    @property
    def celery(self):
        from portal.async_celery import celery_app
        return celery_app

    @property
    def storage(self):
        from portal import storage
        return storage

    def reset_task_id(self, state=None):
        if state is not None:
            self.submission.state = state

        self.submission.async_task_id = None
        general.write_entity(self.submission)

    def submission_enqueue_ended(self):
        log.info(f"[ASYNC] Submission enqueue ended: {self.submission}")
        self.reset_task_id(state=SubmissionState.QUEUED)

    def revoke_enqueue_task(self):
        log.info(f'[ASYNC] Submission storage upload cancelled {self.submission}')
        task_id = self.submission.process_task_id
        if task_id:
            self.celery.control.revoke(task_id=task_id, terminate=True)
            self.reset_task_id(state=SubmissionState.CANCELLED)

    def get_delay_for_submission(self):
        log.debug(f"[ASYNC] Submission delay: {self.submission}")
        time_wait = self.project.config.submissions_cancellation_period
        return time_wait

    def dispatch_submission_processing(self):
        delay = self.get_delay_for_submission()
        args = (self.submission.id, self.params)
        from .tasks import start_processing_submission
        start_processing_submission.apply_async(args=args, countdown=delay)

    def submission_store_ended(self, version: str):
        log.info(f"[ASYNC] Submission preparation ended: {self.submission}")
        self.submission.source_hash = version
        self.reset_task_id(state=SubmissionState.READY)

    def revoke_storage_task(self):
        log.info(f'[ASYNC] Submission storage upload cancelled {self.submission}')
        task_id = self.submission.storage_task_id
        if task_id:
            self.celery.control.revoke(task_id=task_id, terminate=True)
            self.reset_task_id(state=SubmissionState.CANCELLED)
        # TODO: Dispatch cleanup task

    def download_submission(self):
        log.info(f"[ASYNC] Uploading submission: {self.submission} with {self.params}")
        updated_entity: UploadedEntity = self.storage. \
            submissions.create(entity_id=self.submission.id, **self.params)
        self.submission_store_ended(version=updated_entity.version)

    def clone(self, target):
        log.info(f"[ASYNC] Cloning submission: {self.submission} to {target}")
        self.storage.submissions.clone(self.submission.id, target.id)
        self.submission_store_ended(version=self.submission.source_hash)

    def send_to_worker(self):
        # TODO: implement processing
        log.info(f"[ASYNC] Sending submission to worker: {self.submission}")
        if self.params is not None:
            print("Not implemented")

    def upload_result(self, path):
        log.info(
            f"[ASYNC] Uploading result for the submission {self.submission.id} with {self.params}")
        self.storage.results.create(entity_id=self.submission.id, **self.params)
        Path(path).unlink()
        self.reset_task_id(SubmissionState.FINISHED)

    def process_submission(self):
        log.info(f"[ASYNC] Processing submission: {self.submission}")
        self.download_submission()
        self.dispatch_submission_processing()
