"""
Roles service
"""

import logging
from typing import Union, List

from portal.database.models import Course, Role, RolePermissions, User, Client
from portal.service.errors import ForbiddenError, PortalAPIError
from portal.service.filters import filter_roles_from_course
from portal.service.general import delete_entity, get_new_name, update_entity, write_entity, find_client
from portal.service.permissions import check_client
from portal.service.users import find_users

log = logging.getLogger(__name__)


def copy_role(source: Role, target: Course, with_clients: str) -> Role:
    """Copies role to the target course
    Args:
        source(Role): Source role resource
        target(Course): Target Course resource
        with_clients(str): With users flag

    Returns(Role): Copied role
    """
    log.info(
        f"[COPY] Role: {source.id} to course {target.id} with config: {with_clients}")
    new_name = get_new_name(source, target)
    new_role = Role(target, codename=new_name)
    new_role.set_permissions(**vars(source.permissions))
    new_role.description = source.description
    new_role.name = source.name
    if with_clients == "with_users":
        for clients in source.clients:
            clients.roles.append(new_role)
    return new_role


def delete_role(role: Role):
    """Deletes role for the DB
    Args:
        role(Role): Role instance
    """
    log.info(f"[DELETE] Role: {role.id} ({role.name})")
    delete_entity(role)


def update_role(role: Role, data: dict) -> Role:
    """Updates role based on data
    Args:
        role(Role): Role instance
        data(dict): Data for role update
    Returns(Role): Updated role
    """
    log.info(f"[UPDATE] Role {role.id} to {role}.")
    return __set_role_data(role, data)


def __set_role_data(role: Role, data: dict):
    return update_entity(role, data, allowed=['name', 'description', 'codename'])


def create_role(course: Course, **data) -> Role:
    """Creates new role
    Args:
        course(Course): Course instance

    Keyword Args:
        name(str): Name of the role
        description(str): Description of the role
        codename(str): Codename of the role

    Returns(Role): New role instance
    """
    new_role = Role(course=course)
    __set_role_data(role=new_role, data=data)
    log.info(f"[CREATE] Role for course {course.id}: {new_role}")
    return new_role


def update_role_permissions(role: Role, permissions: dict) -> RolePermissions:
    """Updates role permissions
    Args:
        role(Role): Role instance
        permissions(dict): Permissions dictionary

    Returns(RolePermissions): Role Permissions instance
    """
    role.set_permissions(**permissions)
    write_entity(role)
    log.info(f"[UPDATE] Permissions for role {role.id} "
             f"in course {role.course.id}: {permissions}.")
    return role.permissions


def find_clients(ids: List[str]) -> List[Client]:
    """Finds all clients based on their ids

        Args:
            ids(List[str]): List of client ids

        Returns(List[Client]): List of clients
        """
    return [find_client(i) for i in ids]


def update_clients_membership_in_role(role: Role, data: dict) -> Role:
    """Updates client membership in the role
    Args:
        role(Role): Role instance
        data(dict): Data provided to update role membership

    Returns(Role): Updated role
    """
    clients_old = set(role.clients)
    clients_add = data.get('add')
    clients_remove = data.get('remove')
    clients = set()
    if clients_add:
        to_add = find_clients(data['add'])
        clients = clients_old.union(to_add)
    if clients_remove:
        to_remove = find_clients(data['remove'])
        clients = clients_old.difference(to_remove)
    role.clients = list(clients)
    write_entity(role)
    log.info(f"[UPDATE] Clients in Role {role.id}: {data}.")
    return role


def add_single_client_to_role(role: Role, client: Client):
    """Adds single client to the role
    Args:
        role(Role): Role instance
        client(Client): Client instance
    Returns:
    """
    if client not in role.clients:
        role.clients.append(client)
        write_entity(role)
        log.info(f"[ADD] Client {client.id} to role "
                 f"{role.id} in course {role.course.id}.")
    else:
        log.info(f"[ADD] Client {client.id} is already "
                 f"in role {role.id} in course {role.course.id}: no change.")


def remove_single_client_from_role(role: Role, client: Client) -> Role:
    """Removes single client from the role
    Args:
        role(Role): Role instance
        client(Client): Client instance
    Returns(Role): Updated role
    """
    course = role.course
    try:
        role.clients.remove(client)
        write_entity(role)
    except ValueError:
        raise PortalAPIError(400, message=f"Could not remove client {client.id} "
                                          f"from role {role.id} in course {course.id}: "
                                          f"role does not contain client.")
    log.info(
        f"[REMOVE] Client {client.id} from role {role.id} in course {course.id}.")
    return role


def list_roles(course: Course, client: Client) -> List[Role]:
    """List of all roles
    Args:
        course(Course): Course instance
        client(Client): Client instance

    Returns(list): List of roles

    """
    if check_client(client=client, course=course,
                    permissions=['view_course_full']):
        return course.roles
    elif check_client(client=client, course=course, permissions=['view_course_limited']):
        return filter_roles_from_course(course=course, client=client)
    raise ForbiddenError(uid=client.id)
