"""
Submissions service
"""

import logging
from pathlib import Path
from typing import Union

import flask
from celery.result import AsyncResult
from flask import request
from storage import entities
from werkzeug.utils import secure_filename

from portal import storage
from portal.async_celery import tasks
from portal.async_celery.enqueue import revoke_enqueue_task
from portal.async_celery.tasks import clone_submission_files, process_submission
from portal.database.models import Project, Submission, SubmissionState, User, Worker
from portal.service import errors
from portal.service.errors import ForbiddenError
from portal.service.general import delete_entity, write_entity

log = logging.getLogger(__name__)


def create_submission(user: User, project: Project, file_params: dict,
                      project_params_string: str) -> Submission:
    """Creates a new submission in the database and downloads its files.

    Zip and git sources are processed differently.

    Args:
        user(User): the submission's author
        project(Project): the project the submission falls under
        file_params(dict): parameters for Storage - how and where to download files
        project_params_string(str): custom parameters specified by
            project with their values as a JSON-encoded string

    Returns(Submission): the created submission entity

    """
    # adding review at submission create is not supported
    new_submission = Submission(
        user=user, project=project, parameters=project_params_string)
    write_entity(new_submission)

    # download relevant files into Storage
    process_new_submission(new_submission, file_params)
    return new_submission


def process_new_submission(new_submission: Submission, file_params: dict) -> AsyncResult:
    project = new_submission.project
    file_params['whitelist'] = project.config.file_whitelist
    source = file_params['source']
    if source['type'] == 'zip':
        source['path'] = "todo"  # TODO: download zip

    task = create_process_submission_task(file_params, new_submission)
    return task


def create_process_submission_task(file_params, new_submission) -> AsyncResult:
    result: AsyncResult = process_submission.delay(
        new_submission.id, file_params)
    new_submission.storage_task_id = result.task_id
    return result


def copy_submission(source: Submission, note: str) -> Submission:
    """Copies a submission. Used at resubmitting

    Args:
        source(Submission):
        note(str): Note for the submission. Typically contains a reason for the resubmit.

    Returns(Submission): Copied submission
    """
    log.info(f"[COPY] Submission {source.id} to "
             f"same project {source.project.id}")
    new_submission = Submission(user=source.user, project=source.project,
                                parameters=source.parameters)
    new_submission.note = note
    write_entity(new_submission)
    clone_submission_files.delay(
        source_id=source.id, target_id=new_submission.id)
    return new_submission


def delete_submission(submission: Submission):
    """Deletes a submission

    Args:
        submission(Submission): Submission instance
    """
    log.info(f"[DELETE] Submission {submission.id}")
    # TODO: dispatch action to clean the storage
    delete_entity(submission)


def update_submission_state(client: Union[User, Worker], submission: Submission,
                            data: dict) -> Submission:
    """Updates submission's state

    Args:
        client(User): Client attempting the update
        submission(Submission): Submission instance to update
        data(dict): should contain one key, 'state' with the new value

    Returns(Submission): Updated submission
    """
    new_state = data['state']
    if isinstance(client, User) and new_state != SubmissionState.CANCELLED:
        raise ForbiddenError(uid=client.id,
                             note=f"User {client.id} cannot update "
                                  f"state to other than CANCELLED.")
    submission.change_state(new_state)
    write_entity(submission)
    log.info(f"[UPDATE] Submission state {submission.id} "
             f"by {client.id} to {submission.state}")
    return submission


def cancel_submission(submission: Submission):
    """Cancels the submission
    Args:
        submission(Submission): Submission instance

    """
    submission.state = SubmissionState.CANCELLED
    revoke_enqueue_task(submission=submission)
    revoke_storage_task(submission=submission)
    write_entity(submission)


def send_zip(storage_submission: entities.Submission):
    path = storage_submission.zip_path
    if not path.exists():
        raise errors.PortalAPIError(400, f"Requested path does not exist: {path}")
    log.debug(f"[SEND] Sending zip file for submission ({storage_submission.entity_id}): {path}")
    return flask.send_file(str(path), attachment_filename=path.name)


def send_selected_file(storage_submission: entities.Submission, path_query: str):
    path_query = Path(path_query)
    path = (storage_submission.path / path_query)
    path = path.absolute()
    if not path.exists():
        raise errors.PortalAPIError(400, f"Requested path does not exist: {path}")
    log.debug(f"[SEND] Sending file for submission ({storage_submission.entity_id}): {path}")
    return flask.send_file(str(path), attachment_filename=path.name)


def send_file_or_zip(storage_entity):
    path_query = request.args.get('path')
    if path_query is None:
        return send_zip(storage_entity)
    return send_selected_file(storage_entity, path_query)


def send_files_tree(storage_entity):
    tree = storage_entity.tree()
    log.debug(f"[TREE] Tree for the storage entity {storage_entity.entity_id}: {tree} ")
    return tree


def upload_file_is_allowed(file):
    extension = Path(file.filename).suffix
    log.debug(f"[ZIP] Extension for {file.filename}: {extension}")
    if not extension == '.zip':
        raise errors.PortalAPIError(
            400,
            f"File with extension \"{extension}\" is not allowed: #{file.filename}"
        )


def upload_files_to_storage(file):
    filename = secure_filename(file.filename)
    uploads = storage.results.workspace_path / 'uploads'
    if not uploads.exists():
        uploads.mkdir(parents=True)
    path = uploads / filename
    file.save(str(path))
    log.info(f"[UPLOAD] Uploading file to {filename} to {path}")
    return path


def upload_results_to_storage(submission):
    path = get_upload_file_path()
    task = tasks.upload_results_to_storage.delay(submission.id, path=str(path))
    return task


def get_upload_file_path():
    file = request.files['file']
    upload_file_is_allowed(file)
    path = upload_files_to_storage(file)
    return path
